// Flexconnect
// Copyright 2016,2017 Hardy Lau, DL1GLH
//
// v171213-0.93 Alpha
//
// Standalone interface for interconnection of
// + Flexradio 6300/6500/6700
// + S.P.E. Expert 2K-FA
// + Ultrabeam dynamic antenna
//
// Running on Banana Pro
// This part (process) does connect to Ultrabeam
//

/*
 Copyright 2016 Hardy Lau, DL1GLH

 Redistribution and use in source and binary forms, with or without modification,
 are permitted provided that the following conditions are met:

 1. Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.

 3. Neither the name of the copyright holder nor the names of its contributors
    may be used to endorse or promote products derived from this software without
    specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <stdio.h>
#include <stdlib.h>
#include <netdb.h>
#include <netinet/in.h>
#define _GNU_SOURCE
#include <string.h>
#include <sys/types.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <termios.h>
#include <unistd.h>
#include <math.h>
#include "connect.h"
#include <libgen.h>

#define DEBUG
#undef DEBUG

//Retract elements
//
int retract_ub( int tty)
{
 unsigned char checksum = 0x55;
 unsigned char buffer[1];
 unsigned char seq = 0x55;

 buffer[0] = STX;
 write( tty, buffer, 1);

 buffer[0] = seq;
 write( tty, buffer, 1);
 checksum ^= seq;
 checksum++;

 buffer[0] = 0x02;
 write( tty, buffer, 1);
 checksum ^= 0x02;
 checksum++;
 
 switch( checksum) {
   case STX:
   case ETX:
   case DLE: buffer[0] = DLE;
              write( tty, buffer, 1);
              buffer[0] = checksum & 0x7f;
              write( tty, buffer, 1);
              break;
   default:   buffer[0] = checksum;
              write( tty, buffer, 1);
 }

 buffer[0] = ETX;
 write( tty, buffer, 1);
}

//send ub50 new frequency
//
int send_ub_freq( int tty, double frequency, int dir)
{
 unsigned char checksum = 0x55;
 unsigned char buffer[1];
 unsigned char seq = 0x44;
 unsigned int f;
 unsigned char fl, fh;

 buffer[0] = STX;
 write( tty, buffer, 1);

 buffer[0] = seq;
 write( tty, buffer, 1);
 checksum ^= seq;
 checksum++;

 buffer[0] = 0x03;
 write( tty, buffer, 1);
 checksum ^= 0x03;
 checksum++;

 f = (unsigned int)(frequency * 1000);
 fh = f >> 8; 
 fl = f & 255;

 switch( fl) {
   case STX:
   case ETX:
   case DLE: buffer[0] = DLE;
              write( tty, buffer, 1); 
              buffer[0] = fl & 0x7f;
              write( tty, buffer, 1);
              break;
   default:   buffer[0] = fl;
              write( tty, buffer, 1); 
 }
 checksum ^= fl;
 checksum++;

 switch( fh) {
   case STX:
   case ETX:
   case DLE: buffer[0] = DLE;
              write( tty, buffer, 1);
              buffer[0] = fh & 0x7f;
              write( tty, buffer, 1);
              break;
   default:   buffer[0] = fh;
              write( tty, buffer, 1);
 }
 checksum ^= fh;
 checksum++;

 if( dir != NODIR) { 
   buffer[0] = dir;
   write( tty, buffer, 1);
   checksum ^= dir;
   checksum++;
 }

 switch( checksum) {
   case STX:
   case ETX:
   case DLE: buffer[0] = DLE;
              write( tty, buffer, 1);
              buffer[0] = checksum & 0x7f;
              write( tty, buffer, 1);
              break;
   default:   buffer[0] = checksum;
              write( tty, buffer, 1);
 }

 buffer[0] = ETX;
 write( tty, buffer, 1);
}

//Querry status
//
int request_ub_status( int tty)
{
 unsigned char checksum = 0x55;
 unsigned char buffer[1];
 unsigned char seq = 0x4;

 buffer[0] = STX;
 write( tty, buffer, 1);

 buffer[0] = seq;
 write( tty, buffer, 1);
 checksum ^= seq;
 checksum++;

 buffer[0] = 0x01;
 write( tty, buffer, 1);
 checksum ^= 0x01;
 checksum++;

 switch( checksum) {
   case STX:
   case ETX:
   case DLE: buffer[0] = DLE;
              write( tty, buffer, 1);
              buffer[0] = checksum & 0x7f;
              write( tty, buffer, 1);
              break;
   default:   buffer[0] = checksum;
              write( tty, buffer, 1);
 }

 buffer[0] = ETX;
 write( tty, buffer, 1);
}

//Read_char with timeout from serial port
//
int read_char( int tty)
{
 char buffer[2];
 fd_set rfds;
 struct timeval tv;
 int retval;

 FD_ZERO( &rfds);
 FD_SET( tty, &rfds);

 /* Wait max 0,8 seconds. */
 tv.tv_sec = 0;
 tv.tv_usec = 800000L;

 retval = select( tty + 1, &rfds, NULL, NULL, &tv);

 /* Wait with timeout */
 if( retval > 0) {
   read( tty, buffer, 1);
   return( buffer[0] & 255);
 }
 else { /* Timeout */
   return( -2);
 }
}

//Read answer from ultrabeam
//
int read_ub( int tty, unsigned char *buffer)
{
 unsigned char checksum, checksum_old, lastbyte, c;
 int in_dle = 0;
 int in_rx = 0;
 int i = 0;
 int x;

 do {
   x = read_char(tty);
   if( x < 0)
     return( x);

   c = (unsigned char)(x % 256);

   switch( c) {
     case STX:  checksum = 0x55;
                int i = 0;
                buffer[i++] = STX;
                in_rx = 1;
                break;
     case DLE:  in_dle = 1;
                break;
     case ETX:  buffer[i++] = ETX;
                if( lastbyte == checksum_old)
                  return( 0);
                else
                  return( -3);
                break;
     default:   if( in_rx) {
                  checksum_old = checksum;
                  if( in_dle) {
                    in_dle = 0;
                    c += 128;
                  }
                  lastbyte = c; 
                  buffer[i++] = c; 
                  checksum ^= c;
                  checksum++;
                }
   }
 } while( i < 128);
 return( -1);
}


//read a file /run/flexconnect/frequency-%1.6lf
//
double getfrequency()
{
 DIR *dir;
 struct dirent *dp;
 char *file_name;
 char file_name_full_path[200];
 int fd;
 double frequency = 0.0;
 int x = 30;

 dir = opendir( "/run/flexconnect");
 if( dir < 0) return( 0.0); 
 
 while ((dp=readdir(dir)) != NULL && x--) {
   file_name = dp->d_name;
   if( strstr( file_name, "frequency-")) {
     sscanf( file_name, "frequency-%lf", &frequency);
   }
 }
 closedir( dir);
 return( frequency);
}


//Clear UB status file  /run/flexconnect/ub-status-<PREFIX>
//
int clear_ub_status( char *prefix)
{
 DIR *dir;
 struct dirent *dp;
 char *file_name;
 char file_name_full_path[200];

 //Delete all status files
 dir = opendir( "/run/flexconnect");
 while ((dp=readdir(dir)) != NULL) {
   file_name = dp->d_name;
   if( strstr( file_name, prefix)) {
     strcpy( file_name_full_path, "/run/flexconnect/");
     strcat( file_name_full_path, file_name);
     //Delete all status files
     unlink( file_name_full_path);
   }
 }
 closedir( dir);
}

//Get next UB action request (if available)
//
char* getubrequest()
{
 DIR *dir;
 struct dirent *dp;
 char *file_name;
 static char returnstring[200];
 int fd;

 strcpy( returnstring, "");

 dir = opendir( "/run/flexconnect");
 if( dir < 0) return( returnstring);

 while ((dp=readdir(dir)) != NULL) {
   file_name = dp->d_name;
   if( strstr( file_name, "ub-request-")) {
     strcpy( returnstring, file_name);
     break;
   }
 }
 closedir( dir);

 clear_ub_status( "ub-request-");
 return( returnstring);
}

//Write UB status file /run/flexconnect/ub-status-%s
//
int write_ub_status( unsigned char* buffer)
{
 DIR *dir;
 struct dirent *dp;
 char *file_name;
 char file_name_full_path[200];
 char file_name_full_path_new[200];
 int fd;

 //Create dir - doesn't matter if already exist
 mkdir( "/run/flexconnect", 0777);

 clear_ub_status( "ub-status-");

 //Create file with status information
 sprintf( file_name_full_path_new, "/run/flexconnect/ub-status-%02x-%02x-%02x-%02x-%02x-%02x-%02x-%02x-%02x-%02x-%02x-%02x-%02x-%02x-%02x-%02x", buffer[0], buffer[1], buffer[2], buffer[3], buffer[4], buffer[5], buffer[6], buffer[7], 
      buffer[8], buffer[9], buffer[10], buffer[11],buffer[12], buffer[13], buffer[14], buffer[15], buffer[16], buffer[17]);
 fd = creat( file_name_full_path_new, 0777);
 close( fd);
}

//Main
//
int main( int argc, char *argv[])
{
  int i;
  int x;
  int tty;
  int baud = B19200;
  struct termios newtio;
  unsigned char buffer[255];
  double frequency, oldfrequency;
  double fs, fl = 0.0, fu = 0.0; 
  int tuning = 0; 
  int ubrequestactive = 0;
  int ubdirection = NODIR;
  int ubretract = 0;

  if (argc < 2) {
     fprintf(stderr,"usage %s tty_device\n", argv[0]);
     exit(0);
  }
  
  tty = open( argv[1], O_RDWR | O_NOCTTY ); 
  if (tty <0) { 
    perror(argv[1]); 
    exit(-1); 
  }

#ifdef DEBUG
  printf( "UBConnect v0.7 - (c)2016 Hardy Lau - DL1GLH\n");
#endif

  bzero(&newtio, sizeof(newtio)); /* clear struct for new port settings */
        
  newtio.c_cflag = CS8 | CLOCAL | CREAD;
  cfsetospeed( &newtio, B19200);
  cfsetispeed( &newtio, B19200);
  //newtio.c_iflag = IGNPAR | ICRNL;
  newtio.c_iflag = IGNBRK | INPCK;
  newtio.c_oflag = 0;
  newtio.c_lflag = 0; // ICANON;
  tcflush(tty, TCIFLUSH);
  tcsetattr(tty,TCSANOW,&newtio); 

  wiringPiSetup () ;
  pinMode(TX_INHIBIT_FLEX, OUTPUT);
  pinMode(PTT_FLEX, INPUT);
  pinMode(PTT_ENABLE_FLEX, OUTPUT);
  pinMode(CONTROL_UB, OUTPUT);

  digitalWrite(PTT_ENABLE_FLEX, HIGH);

  while( 1) {
    frequency = getfrequency();
    
    //Only if PTT is not active
    if( digitalRead(PTT_FLEX)) {

      //Only if in frequency range of Ultrabeam
      if( (frequency >= ULTRABEAM_LOW_FREQUENCY_EDGE) && (frequency <= ULTRABEAM_HIGH_FREQUENCY_EDGE)) {
        //We control ultrabeam
        digitalWrite (CONTROL_UB, HIGH);
        if( fl == 0.0 && fu == 0.0) {
          do {
            x = read_ub(tty, buffer);
          } while( x >= 0);
        }

        //Any request from webserver
        strcpy( buffer, getubrequest());
        if( strlen(buffer)) {
          ubrequestactive = 1;
          if( strstr( buffer, "ub-request-ubnormal")) ubdirection = DIRNORM;
          if( strstr( buffer, "ub-request-ubreverse")) ubdirection = DIRREV;
          if( strstr( buffer, "ub-request-ubbimode")) ubdirection = DIRBI;
          if( strstr( buffer, "ub-request-ubretract")) ubretract = 1;
        }     

        //Step out of 25kHz segment
        if( (frequency > fu) || (frequency < fl) || ubrequestactive) {
      
          ubrequestactive = 0;
 
          //New frequency in step size
          fs = 25.0 * round(frequency * 1000.0 / 25.0);

          //Protect Ultrabeam while tuning
          digitalWrite (TX_INHIBIT_FLEX,  HIGH);
          digitalWrite(PTT_ENABLE_FLEX, LOW);
#ifdef DEBUG
          printf("TUNING\n"); fflush(stdout);
#endif
          tuning = 1;
          fu = (fs + 14.5) / 1000.0;
          fl = (fs - 14.5) / 1000.0;
 
#ifdef DEBUG
          printf( "%lf Mc - %lf  > %lf <  %lf\n", frequency, fl, fs/1000.0, fu);
#endif 
          do { 
            send_ub_freq( tty, fs / 1000.0, ubdirection);   
            x = read_ub( tty, buffer);
#ifdef DEBUG
            for( i = 0; i < 5; i++) { printf( "%02x ", (unsigned char)buffer[i]); } printf( "\n");
#endif
            usleep( 50000L);

          } while( x != 0 || buffer[1] != 0x44 || buffer[2] != 0);
          ubdirection = NODIR;
        }
      } else { //No PTT but out of frequency of ultrabeam
        //No frequency, no coltrol of ultrabeam
        if( ubretract) {
#ifdef DEBUG
          printf( "Retracting UltraBeam...");
#endif
          retract_ub( tty);
          sleep( 1);
          ubretract = 0;
        }
        digitalWrite (CONTROL_UB, LOW);
        fl = 0.0;
        fu = 0.0;
        clear_ub_status( "ub-status-");
      }
    }

    //Here down PTT may be active !!! 
   
    //Request status
    request_ub_status( tty);
    x = read_ub( tty, buffer);
#ifdef DEBUG
    for( i = 0; i < 16; i++) { printf( "%02x ", (unsigned char)buffer[i]); } printf( "\n");
#endif
    if(  x == 0 && buffer[1] == 0x4) {
      if( buffer[12] == 0) { //Ultrabeam is ready
        //Tuning ready
        digitalWrite(TX_INHIBIT_FLEX, LOW);
	digitalWrite(PTT_ENABLE_FLEX, HIGH);
        if( tuning) { 
#ifdef DEBUG
          printf("READY\n"); fflush(stdout);
#endif
          tuning = 0;
        }
      } else { //Ultrabeam is tuning!!!
        //Protect Ultrabeam while tuning
        digitalWrite (TX_INHIBIT_FLEX, HIGH);
	digitalWrite(PTT_ENABLE_FLEX, LOW);
        if( tuning == 0) {
#ifdef DEBUG
          printf("PROTECT\n"); fflush(stdout);
#endif
          tuning = 1;
        }
      }
      write_ub_status( buffer); 
    }
    usleep(100000L);
  }
}
